UPDATE GUN_DATA SET OG_MAKE = 'H&K' WHERE MAKE = 'HK';
UPDATE GUN_DATA SET OG_MAKE = 'HECKLER & KOCH' WHERE MAKE = 'H&K';
UPDATE GUN_DATA SET OG_MAKE = 'SPRINGFIELD ARMORY' WHERE MAKE = 'SPRINGFIELD';
UPDATE GUN_DATA SET OG_MAKE = 'AERO PRECISION' WHERE MAKE = 'AERO';
UPDATE GUN_DATA SET OG_MAKE = 'SMITH & WESSON' WHERE MAKE = 'SW';
UPDATE GUN_DATA SET OG_MAKE = 'HARRINGTON & RICHARDSON' WHERE MAKE = 'H&R';
UPDATE GUN_DATA SET OG_MAKE = 'HENRY REPEATING ARMS' WHERE MAKE = 'HENRY';
UPDATE GUN_DATA SET OG_MAKE = 'SAVAGE' WHERE MAKE = 'SAVAGE / ANSCHUTZ';
UPDATE GUN_DATA SET OG_MAKE = 'STAG ARMS' WHERE MAKE = 'STAG';
UPDATE GUN_DATA SET OG_MAKE = 'CZ' WHERE MAKE = 'CZ-USA';
UPDATE GUN_DATA SET OG_MAKE = 'ANDERSON MANUFACTURING' WHERE MAKE = 'ANDERSON';
UPDATE GUN_DATA SET OG_MAKE = 'ANDERSON MANUFACTURING' WHERE MAKE = 'ANDERSON MFG';
UPDATE GUN_DATA SET OG_MAKE = 'ROCK ISLAND ARMORY' WHERE MAKE = 'ROCK ISLAND';
UPDATE GUN_DATA SET OG_MAKE = 'SAVAGE' WHERE MAKE = 'SAVAGE ARMS';
UPDATE GUN_DATA SET OG_MAKE = 'FN' WHERE MAKE = 'FN AMERICA';
UPDATE GUN_DATA SET OG_MAKE = 'WINDHAM WEAPONRY' WHERE MAKE = 'WINDHAM';
UPDATE GUN_DATA SET OG_MAKE = 'WINDHAM WEAPONRY' WHERE MAKE = 'WINDHAM ARMS';
UPDATE GUN_DATA SET OG_MAKE = 'WINDHAM WEAPONRY' WHERE MAKE = 'WINDHAM ARMAMENT';
UPDATE GUN_DATA SET OG_MAKE = 'WINDHAM WEAPONRY' WHERE MAKE = 'WINDHAM WEAPONRY -USA';
UPDATE GUN_DATA SET OG_MAKE = 'WINDHAM WEAPONRY' WHERE MAKE = 'WINDHAM WEAPONRY (LE SALE';
UPDATE GUN_DATA SET OG_MAKE = 'WINDHAM WEAPONRY' WHERE MAKE = 'WINDHAM ARMS WEAPONRY';
UPDATE GUN_DATA SET OG_MAKE = 'WINDHAM WEAPONRY' WHERE MAKE = 'WINDHAM ARMORY';
UPDATE GUN_DATA SET OG_MAKE = 'SMITH & WESSON' WHERE MAKE LIKE '%SMITH&%';
UPDATE GUN_DATA SET OG_MAKE = 'SMITH & WESSON' WHERE MAKE LIKE '%SMITH &%';
UPDATE GUN_DATA SET OG_MAKE = 'SMITH & WESSON' WHERE MAKE LIKE '%SMITH W%';
UPDATE GUN_DATA SET OG_MAKE = 'SMITH & WESSON' WHERE MAKE LIKE '%SMITH+%';
UPDATE GUN_DATA SET OG_MAKE = 'SMITH & WESSON' WHERE MAKE LIKE '%SMITHA%';
UPDATE GUN_DATA SET OG_MAKE = 'RUGER' WHERE MAKE LIKE '%RUGER%';
UPDATE GUN_DATA SET OG_MAKE = 'SIG SAUER' WHERE MAKE LIKE '%SIG SAUER%';
UPDATE GUN_DATA SET OG_MAKE = 'SIG SAUER' WHERE MAKE LIKE 'SIG ARMS';
UPDATE GUN_DATA SET OG_MAKE = 'SIG SAUER' WHERE MAKE LIKE 'SIG';
UPDATE GUN_DATA SET OG_MAKE = 'SIG SAUER' WHERE MAKE LIKE 'SIG SWISS';
UPDATE GUN_DATA SET OG_MAKE = 'SIG SAUER' WHERE MAKE LIKE 'SAUER / SIG ARMS';
UPDATE GUN_DATA SET OG_MAKE = 'GLOCK' WHERE MAKE LIKE '%GLOCK%';
UPDATE GUN_DATA SET OG_MAKE = 'REMINGTON' WHERE MAKE LIKE '%REMINGTON%';
UPDATE GUN_DATA SET OG_MAKE = 'MOSSBERG' WHERE MAKE LIKE '%MOSSBERG%';
UPDATE GUN_DATA SET OG_MAKE = 'BERETTA' WHERE MAKE LIKE '%BERETTA%';
UPDATE GUN_DATA SET OG_MAKE = 'SAVAGE' WHERE MAKE LIKE 'SAVAGE ARMS CANADA';
UPDATE GUN_DATA SET OG_MAKE = 'SAVAGE' WHERE MAKE LIKE 'SAVAGE ARMS ONT CAN / SAVAGE ARM';
UPDATE GUN_DATA SET OG_MAKE = 'SAVAGE' WHERE MAKE LIKE 'SAVAGE ARMS / SAVAGE ARMS';
UPDATE GUN_DATA SET OG_MAKE = 'SAVAGE' WHERE MAKE LIKE 'SAVAGE ARMS CANADA, INC / SAVAGE';
UPDATE GUN_DATA SET OG_MAKE = 'SAVAGE' WHERE MAKE LIKE 'SAVAGE ARMS CORP';
UPDATE GUN_DATA SET OG_MAKE = 'SAVAGE' WHERE MAKE LIKE 'AVAGE ARMS ONT CAN';
UPDATE GUN_DATA SET OG_MAKE = 'SAVAGE' WHERE MAKE LIKE 'SAVAGE ARMS ONT CAN';
UPDATE GUN_DATA SET OG_MAKE = 'SAVAGE' WHERE MAKE LIKE 'SAVAGE ARMS INCORPORATED';
UPDATE GUN_DATA SET OG_MAKE = 'SAVAGE' WHERE MAKE LIKE 'SAVAGE ARMS INC. OF CANADA';
UPDATE GUN_DATA SET OG_MAKE = 'SAVAGE' WHERE MAKE LIKE 'SAVAGE ARMS CANADA / SAVAGE ARMS';
UPDATE GUN_DATA SET OG_MAKE = 'SAVAGE' WHERE MAKE LIKE 'SAVAGE ARMS,INC';
UPDATE GUN_DATA SET OG_MAKE = 'SAVAGE' WHERE MAKE LIKE 'SAVAGE ARMS USA';
UPDATE GUN_DATA SET OG_MAKE = 'SAVAGE' WHERE MAKE LIKE 'SAVAGE ARMS INC / NONE';
UPDATE GUN_DATA SET OG_MAKE = 'HENRY REPEATING ARMS' WHERE MAKE LIKE 'HENRY %';
UPDATE GUN_DATA SET OG_MAKE = 'UBERTI' WHERE MAKE LIKE '%UBERTI%';
UPDATE GUN_DATA SET OG_MAKE = 'SPIKES TACTICAL' WHERE MAKE LIKE '%SPIKE%TACTICAL%';
UPDATE GUN_DATA SET OG_MAKE = 'ROCK ISLAND ARMORY' WHERE MAKE LIKE '%ROCK%IS%';
UPDATE GUN_DATA SET OG_MAKE = 'ROCK RIVER ARMS' WHERE MAKE LIKE 'ROCK RIVER%';
UPDATE GUN_DATA SET OG_MAKE = 'ZASTAVA' WHERE MAKE LIKE 'ZASTAVA%';
UPDATE GUN_DATA SET OG_MAKE = 'FN' WHERE MAKE LIKE 'FN%';
UPDATE GUN_DATA SET OG_MAKE = 'SPIKES TACTICAL' WHERE MAKE LIKE 'SPIKE%';
UPDATE GUN_DATA SET OG_MAKE = 'SPRINGFIELD ARMORY' WHERE MAKE LIKE 'SPRINGFIELD%';
UPDATE GUN_DATA SET OG_MAKE = 'CENTURY ARMS' WHERE MAKE LIKE 'CENTURY%';
UPDATE GUN_DATA SET OG_MAKE = 'YANKEE HILL MACHINE' WHERE MAKE LIKE 'YANKEE HILL%';
UPDATE GUN_DATA SET OG_MAKE = 'YANKEE HILL MACHINE' WHERE MAKE LIKE 'YHM%';
UPDATE GUN_DATA SET OG_MAKE = 'YANKEE HILL MACHINE' WHERE MAKE LIKE 'YMH';
UPDATE GUN_DATA SET OG_MAKE = 'YANKEE HILL MACHINE' WHERE MAKE LIKE 'YNAKEE HILL MACHINE CO YHM';
UPDATE GUN_DATA SET OG_MAKE = 'WINDHAM WEAPONRY' WHERE MAKE LIKE 'WINDOM%';
UPDATE GUN_DATA SET OG_MAKE = 'WEATHERBY' WHERE MAKE LIKE 'WEATHERBY%';
UPDATE GUN_DATA SET OG_MAKE = 'WEATHERBY' WHERE MAKE LIKE 'WEATHBY';
UPDATE GUN_DATA SET OG_MAKE = 'CESKA ZBROJOVKA' WHERE MAKE LIKE 'CESK%';
