UPDATE gun_data SET serial = "NONE" WHERE serial LIKE "NSN";
UPDATE gun_data SET serial = "NONE" WHERE serial LIKE "N/A";
UPDATE gun_data SET serial = "NONE" WHERE serial LIKE "NA";
UPDATE gun_data SET serial = "NONE" WHERE serial LIKE "0";
UPDATE gun_data SET serial = "NONE" WHERE serial LIKE "";
UPDATE gun_data SET serial = "NONE" WHERE serial LIKE "1";
UPDATE gun_data SET serial = "NONE" WHERE serial LIKE "2";
UPDATE gun_data SET serial = "NONE" WHERE serial LIKE "NO SERIAL NUMBER";
UPDATE gun_data SET serial = "NONE" WHERE serial LIKE "UNKNOWN";
UPDATE gun_data SET serial = "NONE" WHERE serial LIKE "NO SERIAL";
UPDATE gun_data SET serial = "NONE" WHERE serial LIKE "NOT AVAILABLE";
UPDATE gun_data SET serial = "NONE" WHERE serial LIKE "NO NUMBER";
UPDATE gun_data SET serial = "NONE" WHERE serial LIKE "NONE FOUND";
UPDATE gun_data SET serial = "NONE" WHERE serial LIKE "NONE VISIBLE";
UPDATE gun_data SET serial = "NONE" WHERE serial LIKE "-";
UPDATE gun_data SET serial = "NONE" WHERE serial LIKE "N.S.N";
UPDATE gun_data SET serial = "NONE" WHERE serial LIKE "UNKNOW";
UPDATE gun_data SET serial = "NONE" WHERE serial LIKE "NOT AVAIL";
UPDATE gun_data SET serial = "NONE" WHERE serial LIKE "NVS";
UPDATE gun_data SET serial = "NONE" WHERE serial LIKE "NO SERIAL #";
