#include "game_ui.h"

#if !defined(TILES)

void reinitialize_framebuffer( const bool /*force_invalidate*/ )
{
}

void to_map_font_dim_width( int & )
{
}

void to_map_font_dim_height( int & )
{
}

void to_map_font_dimension( int &, int & )
{
}
void from_map_font_dimension( int &, int & )
{
}
void to_overmap_font_dimension( int &, int & )
{
}

#endif
