from ..write_text import write_text


def parse_tool_quality(json, origin):
    if "name" in json:
        write_text(json["name"], origin,
                   comment="Name of tool quality \"{}\"".format(json["id"]))
